@@ -173,7 +173,7 @@ static int start_ldap_connection(ldap_ctx *o, char *hostnamestr)
 #endif
 
 	x = strchr(hostnamestr, ':');
-	if (!x) {
+	if (!x || (*(x+1) && *(x+1) == '/')) {
 		port = LDAP_PORT;
 	} else {
 		port = atoi(x+1);
@@ -182,8 +182,12 @@ static int start_ldap_connection(ldap_ctx *o, char *hostnamestr)
 
 	o->message_sent = 0;
 	o->message_wait = 0; // XXX
-	o->ldap_con = ldap_init(hostnamestr, port);
-	if (x) *x = ':'; /* egad */
+	if (ldap_is_ldap_url(hostnamestr)) {
+		ldap_initialize(&o->ldap_con, hostnamestr);
+	} else {
+		o->ldap_con = ldap_init(hostnamestr, port);
+		if (x) *x = ':'; /* egad */
+	}
 
 	if (!o->ldap_con)
 		return -1;
@@ -495,6 +499,8 @@ static void do_simple_search(dns_ctx *c, char *q)
 			/* not even going to try... */
 			response_refuse(c);
 			complete_phase(c, '-');
+			mem_free(str(sr));
+			return;
 		}
 		str_addch(sr, j);
 	}
@@ -1851,6 +1857,7 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 	char *dat;
 	list_t lp;
 	int ext_records, selfmatch;
+	int didany = 0;
 
 	/* process NS/SOA search results; possibly rerun */
 	if (!c->message) {
@@ -2131,6 +2138,7 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 					fatal("could not construct address");
 				}
 				response_rfinish(c, RESPONSE_ANSWER);
+				didany++;
 				mem_free(dat);
 			}
 		} else if (c->CNAME) {
@@ -2143,10 +2151,26 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 				response_rfinish(c, RESPONSE_ANSWER);
 				if (!ldapdns.no_additionals)
 					push_subreq(c, dat);
+				didany++;
 				mem_free(dat);
 			}
 		}
 
+		if (!didany && c->NS) {
+			/* if no answers, but we're NS, we give SOA */
+			if (!response_rstart(c, c->request_name_zone, DNS_T_SOA, c->ttl)
+			|| !response_addname(c, ldapdns.self_ns ? ldapdns.self_ns : (c->NS ? c->NS->str : ""))
+			|| !response_addname(c, c->ADM ? c->ADM->str : ldapdns.hostmaster)
+			|| !response_addulong(c, c->serial)
+			|| !response_addulong(c, c->refresh)
+			|| !response_addulong(c, c->retry)
+			|| !response_addulong(c, c->expire)
+			|| !response_addulong(c, c->minimum)) {
+				fatal("could not construct SOA");
+			}
+			response_rfinish(c, RESPONSE_ANSWER);
+		}
+
 		/* also included SOA */
 		while ((dat = list_pop(&c->NS))) {
 			if (!response_rstart(c, c->request_name_zone,
@@ -2198,6 +2222,7 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 					selfmatch = 2;
 				} else if (!ldapdns.no_additionals)
 					push_subreq(c, dat+2);
+				didany++;
 				mem_free(dat);
 			}
 		}
@@ -2213,6 +2238,7 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 					|| !response_addbytes(c, dat, 4)) {
 						fatal("could not construct address");
 					}
+					didany++;
 					response_rfinish(c, RESPONSE_ANSWER);
 					mem_free(dat);
 				}
@@ -2224,12 +2250,27 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 						fatal("could not construct address");
 					}
 					response_rfinish(c, RESPONSE_ANSWER);
+					didany++;
 					if (!ldapdns.no_additionals)
 						push_subreq(c, dat);
 					mem_free(dat);
 				}
 			}
 		}
+		if (!didany && c->NS) {
+			/* if no answers, but we're NS, we give SOA */
+			if (!response_rstart(c, c->request_name_zone, DNS_T_SOA, c->ttl)
+			|| !response_addname(c, ldapdns.self_ns ? ldapdns.self_ns : (c->NS ? c->NS->str : ""))
+			|| !response_addname(c, c->ADM ? c->ADM->str : ldapdns.hostmaster)
+			|| !response_addulong(c, c->serial)
+			|| !response_addulong(c, c->refresh)
+			|| !response_addulong(c, c->retry)
+			|| !response_addulong(c, c->expire)
+			|| !response_addulong(c, c->minimum)) {
+				fatal("could not construct SOA");
+			}
+			response_rfinish(c, RESPONSE_ANSWER);
+		}
 
 		/* also included SOA */
 		while ((dat = list_pop(&c->NS))) {
@@ -2279,10 +2320,26 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 			|| !response_addname(c, dat)) {
 				fatal("could not construct CNAME");
 			}
+			didany++;
 			response_rfinish(c, RESPONSE_ANSWER);
 			mem_free(dat);
 		}
 
+		if (!didany && c->NS) {
+			/* if no answers, but we're NS, we give SOA */
+			if (!response_rstart(c, c->request_name_zone, DNS_T_SOA, c->ttl)
+			|| !response_addname(c, ldapdns.self_ns ? ldapdns.self_ns : (c->NS ? c->NS->str : ""))
+			|| !response_addname(c, c->ADM ? c->ADM->str : ldapdns.hostmaster)
+			|| !response_addulong(c, c->serial)
+			|| !response_addulong(c, c->refresh)
+			|| !response_addulong(c, c->retry)
+			|| !response_addulong(c, c->expire)
+			|| !response_addulong(c, c->minimum)) {
+				fatal("could not construct SOA");
+			}
+			response_rfinish(c, RESPONSE_ANSWER);
+		}
+
 		/* also included SOA */
 		while ((dat = list_pop(&c->NS))) {
 			if (!response_rstart(c, c->request_name_zone,
@@ -2303,9 +2360,24 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 			|| !response_addname(c, dat)) {
 				fatal("could not construct address");
 			}
+			didany++;
 			response_rfinish(c, RESPONSE_ANSWER);
 			mem_free(dat);
 		}
+		if (!didany && c->NS) {
+			/* if no answers, but we're NS, we give SOA */
+			if (!response_rstart(c, c->request_name_zone, DNS_T_SOA, c->ttl)
+			|| !response_addname(c, ldapdns.self_ns ? ldapdns.self_ns : (c->NS ? c->NS->str : ""))
+			|| !response_addname(c, c->ADM ? c->ADM->str : ldapdns.hostmaster)
+			|| !response_addulong(c, c->serial)
+			|| !response_addulong(c, c->refresh)
+			|| !response_addulong(c, c->retry)
+			|| !response_addulong(c, c->expire)
+			|| !response_addulong(c, c->minimum)) {
+				fatal("could not construct SOA");
+			}
+			response_rfinish(c, RESPONSE_ANSWER);
+		}
 
 		/* also included SOA */
 		while ((dat = list_pop(&c->NS))) {
@@ -2343,10 +2415,35 @@ static void ldapdns_process_attrsearch(dns_ctx *c, int no_wild)
 		/* TXT */
 		while ((dat = list_pop(&c->TXT))) {
 			if (!response_rstart(c, c->request_name_alloc, DNS_T_TXT, c->ttl)
-			|| !response_addname(c, dat)) {
+			|| !response_addbytes(c, dat, dns_domain_length(dat)-1)) {
 				fatal("could not construct TXT");
 			}
 			response_rfinish(c, RESPONSE_ANSWER);
+			didany++;
+			mem_free(dat);
+		}
+		if (!didany && c->NS) {
+			/* if no answers, but we're NS, we give SOA */
+			if (!response_rstart(c, c->request_name_zone, DNS_T_SOA, c->ttl)
+			|| !response_addname(c, ldapdns.self_ns ? ldapdns.self_ns : (c->NS ? c->NS->str : ""))
+			|| !response_addname(c, c->ADM ? c->ADM->str : ldapdns.hostmaster)
+			|| !response_addulong(c, c->serial)
+			|| !response_addulong(c, c->refresh)
+			|| !response_addulong(c, c->retry)
+			|| !response_addulong(c, c->expire)
+			|| !response_addulong(c, c->minimum)) {
+				fatal("could not construct SOA");
+			}
+			response_rfinish(c, RESPONSE_ANSWER);
+		}
+		while ((dat = list_pop(&c->NS))) {
+			if (!response_rstart(c, c->request_name_zone, DNS_T_NS, c->ttl)
+			|| !response_addname(c, dat)) {
+				fatal("could not construct NS");
+			}
+			response_rfinish(c, RESPONSE_AUTHORITY);
+			if (!ldapdns.no_additionals && !ldapdns.no_additionals_ns)
+				push_subreq(c, dat);
 			mem_free(dat);
 		}
 	}
@@ -2472,7 +2569,7 @@ static void ldapdns_process_axfrsearch_mid(dns_ctx *c, char *autodn)
 	while ((dat = list_pop(&c->TXT))) {
 		if (!response_axstart(c, 0, str(rn), DNS_T_TXT,
 					DNS_C_IN, c->ttl)
-		|| !response_addname(c, dat)) {
+		|| !response_addbytes(c, dat, dns_domain_length(dat)-1)) {
 			fatal("could not construct TXT");
 		}
 		response_axfinish(c);
@@ -3666,6 +3763,7 @@ int main(int argc, char *argv[])
 	str_init(retbuf);
 	str_copy(retbuf, x);
 	for (h = str(retbuf); h && *h;) { /* on init: np if it infinites */
+		int uri=0;
 		for (x = h; *x && *x != ',' && *x != ' '
 				&& *x != ';' && *x != '\t'; x++);
 		if (*x) {
@@ -3677,18 +3775,24 @@ int main(int argc, char *argv[])
 		p = LDAP_PORT;
 		for (y = h; *y && *y != ':'; y++);
 		if (*y && *y == ':') {
-			*y = 0;
-			y++;
-			p = atoi(y);
-			if (p == -1)
-				p = LDAP_PORT;
+			if (*(y+1) && *(y+1) == '/') { /* detect ldap:// uri */
+				uri=1;
+			} else {
+				*y = 0;
+				y++;
+				p = atoi(y);
+				if (p == -1)
+					p = LDAP_PORT;
+			}
 		}
 
 		str_init(hbuf);
 		str_copy(hbuf, h);
-		str_addch(hbuf, ':');
-		sprintf(spbuf, "%d", p);
-		str_cat(hbuf, spbuf);
+		if (!uri) {
+			str_addch(hbuf, ':');
+			sprintf(spbuf, "%d", p);
+			str_cat(hbuf, spbuf);
+		}
 
 		list_push(&ldapdns.hosts, str(hbuf));
 		h = x;
