@@ -437,12 +437,11 @@ static char inline *__parse_txt(bin_t rrdata)
 	str_t tmp;
 	bin_t retval;
 
-	/* name is in ascii form.... convert to DNS */
+	/* don't dns-encode the TXT segment */
 	bin_init(retval);
-	bin_copy(retval, caddr(rrdata), clen(rrdata));
+	bin_addch(retval, clen(rrdata));
+	bin_cat(retval, caddr(rrdata), clen(rrdata));
 	bin_0(retval);
-	name_to_dns_fix(tmp, (char *)caddr(retval), 1);
-	bin_copy(retval, str(tmp), dns_domain_length(str(tmp)));
 
 	return (char *)caddr(retval);
 }
